@@ -4,22 +4,15 @@
 //! The rest of the crate is used to obtain DICOM element headers and values.
 //! At this level, headers and values are treated as tokens which can be used
 //! to form a syntax tree of a full data set.
-use crate::marker::DicomElementMarker;
-use crate::stateful::decode::{
-    DynStatefulDecoder, Error as DecoderError, StatefulDecode, StatefulDecoder,
-};
-use crate::util::ReadSeek;
-use dicom_core::dictionary::DataDictionary;
+use crate::stateful::decode::{DynStatefulDecoder, Error as DecoderError, StatefulDecode};
 use dicom_core::header::{DataElementHeader, Header, Length, SequenceItemHeader};
 use dicom_core::{PrimitiveValue, Tag, VR};
-use dicom_dictionary_std::StandardDataDictionary;
 use dicom_encoding::text::SpecificCharacterSet;
 use dicom_encoding::transfer_syntax::TransferSyntax;
 use snafu::{Backtrace, ResultExt, Snafu};
 use std::cmp::Ordering;
 use std::io::Read;
 use std::iter::Iterator;
-use std::marker::PhantomData;
 
 use super::{DataToken, SeqTokenType};
 
@@ -72,6 +65,8 @@ pub enum Error {
     },
     #[snafu(display("Unexpected item tag {} while reading element header", tag))]
     UnexpectedItemTag { tag: Tag, backtrace: Backtrace },
+    /// Undefined pixel item length
+    UndefinedItemLength,
 }
 
 pub type Result<T> = std::result::Result<T, Error>;
@@ -125,33 +120,48 @@ pub enum ValueReadStrategy {
     Raw,
 }
 
-/// The set of options for the data set reader.
-#[derive(Debug, Copy, Clone, Eq, Hash, PartialEq)]
-#[non_exhaustive]
-pub struct DataSetReaderOptions {
-    pub value_read: ValueReadStrategy,
+impl Default for ValueReadStrategy {
+    fn default() -> Self {
+        ValueReadStrategy::Preserved
+    }
 }
 
-impl Default for DataSetReaderOptions {
-    fn default() -> Self {
-        DataSetReaderOptions {
-            value_read: ValueReadStrategy::Preserved,
-        }
+/// The set of options for the data set reader.
+#[derive(Debug, Default, Copy, Clone, Eq, Hash, PartialEq)]
+#[non_exhaustive]
+pub struct DataSetReaderOptions {
+    /// the value reading strategy
+    pub value_read: ValueReadStrategy,
+    /// the position of the reader as received at building time
+    pub base_offset: u64,
+}
+
+impl DataSetReaderOptions {
+    /// Replace the value reading strategy of the options.
+    pub fn value_read(mut self, value_read: ValueReadStrategy) -> Self {
+        self.value_read = value_read;
+        self
+    }
+    /// Replace the base reader offset of the options.
+    pub fn base_offset(mut self, base_offset: u64) -> Self {
+        self.base_offset = base_offset;
+        self
     }
 }
 
 /// A higher-level reader for retrieving structure in a DICOM data set from an
 /// arbitrary data source.
 #[derive(Debug)]
-pub struct DataSetReader<S, D> {
+pub struct DataSetReader<S> {
     /// the stateful decoder
     parser: S,
-    /// the data attribute dictionary
-    dict: D,
     /// the options of this reader
     options: DataSetReaderOptions,
-    /// whether the reader is expecting an item next (or a sequence delimiter)
+    /// whether the reader is expecting an item header next (or a sequence delimiter)
     in_sequence: bool,
+    /// whether the reader is expecting the first item value of a pixel sequence next
+    /// (offset table)
+    offset_table_next: bool,
     /// whether a check for a sequence or item delimitation is pending
     delimiter_check_pending: bool,
     /// a stack of delimiters
@@ -160,88 +170,76 @@ pub struct DataSetReader<S, D> {
     hard_break: bool,
     /// last decoded header
     last_header: Option<DataElementHeader>,
-    /// Whether to expect a raw value next, and how many bytes long
-    raw_value_length: Option<u32>,
 }
 
-impl<'s> DataSetReader<DynStatefulDecoder<'s>, StandardDataDictionary> {
+impl<R> DataSetReader<DynStatefulDecoder<R>> {
     /// Creates a new iterator with the given random access source,
     /// while considering the given transfer syntax and specific character set.
-    #[deprecated]
-    pub fn new_with<S: 's>(source: S, ts: &TransferSyntax, cs: SpecificCharacterSet) -> Result<Self>
+    #[deprecated(
+        since = "0.5.0",
+        note = "Use `new_with_ts_cs` or `new_with_ts_cs_options` instead"
+    )]
+    pub fn new_with(source: R, ts: &TransferSyntax, cs: SpecificCharacterSet) -> Result<Self>
     where
-        S: Read,
+        R: Read,
     {
-        let parser = DynStatefulDecoder::new_with(source, ts, cs).context(CreateDecoder)?;
-
-        is_stateful_decode(&parser);
-
-        Ok(DataSetReader {
-            parser,
-            dict: StandardDataDictionary,
-            options: Default::default(),
-            seq_delimiters: Vec::new(),
-            delimiter_check_pending: false,
-            in_sequence: false,
-            hard_break: false,
-            last_header: None,
-            raw_value_length: None,
-        })
+        Self::new_with_ts_cs(source, ts, cs)
     }
-}
 
-impl<'s, D> DataSetReader<DynStatefulDecoder<'s>, D> {
-    /// Creates a new iterator with the given random access source and data dictionary,
-    /// while considering the given transfer syntax and specific character set.
-    pub fn new_with_dictionary<S: 's>(
-        source: S,
-        dict: D,
+    #[inline]
+    pub fn new_with_ts_cs(source: R, ts: &TransferSyntax, cs: SpecificCharacterSet) -> Result<Self>
+    where
+        R: Read,
+    {
+        Self::new_with_ts_cs_options(source, ts, cs, Default::default())
+    }
+
+    /// Create a new iterator with the given stateful decoder and options.
+    pub fn new_with_ts_cs_options(
+        source: R,
         ts: &TransferSyntax,
         cs: SpecificCharacterSet,
         options: DataSetReaderOptions,
     ) -> Result<Self>
     where
-        S: Read,
+        R: Read,
     {
-        let parser = DynStatefulDecoder::new_with(source, ts, cs).context(CreateDecoder)?;
+        let parser = DynStatefulDecoder::new_with(source, ts, cs, 0).context(CreateDecoderSnafu)?;
 
         is_stateful_decode(&parser);
 
         Ok(DataSetReader {
             parser,
-            dict,
             options,
             seq_delimiters: Vec::new(),
             delimiter_check_pending: false,
+            offset_table_next: false,
             in_sequence: false,
             hard_break: false,
             last_header: None,
-            raw_value_length: None,
         })
     }
 }
 
-impl<S> DataSetReader<S, StandardDataDictionary> {
-    /// Create a new iterator with the given parser and options.
+impl<S> DataSetReader<S> {
+    /// Create a new iterator with the given stateful decoder and options.
     pub fn new(decoder: S, options: DataSetReaderOptions) -> Self {
         DataSetReader {
             parser: decoder,
-            dict: StandardDataDictionary,
             options,
             seq_delimiters: Vec::new(),
             delimiter_check_pending: false,
+            offset_table_next: false,
             in_sequence: false,
             hard_break: false,
             last_header: None,
-            raw_value_length: None,
         }
     }
 }
 
-impl<S, D> Iterator for DataSetReader<S, D>
+impl<S> Iterator for DataSetReader<S>
 where
     S: StatefulDecode,
-    D: DataDictionary,
 {
     type Item = Result<DataToken>;
 
@@ -287,19 +285,23 @@ where
                             // closed an item
                             self.seq_delimiters.pop();
                             self.in_sequence = true;
+                            // sequences can end after an item delimiter
+                            self.delimiter_check_pending = true;
                             Some(Ok(DataToken::ItemEnd))
                         }
                         SequenceItemHeader::SequenceDelimiter => {
                             // closed a sequence
                             self.seq_delimiters.pop();
                             self.in_sequence = false;
+                            // items can end after a nested sequence ends
+                            self.delimiter_check_pending = true;
                             Some(Ok(DataToken::SequenceEnd))
                         }
                     }
                 }
                 Err(e) => {
                     self.hard_break = true;
-                    Some(Err(e).context(ReadItemHeader))
+                    Some(Err(e).context(ReadItemHeaderSnafu))
                 }
             }
         } else if let Some(SeqToken {
@@ -309,19 +311,39 @@ where
             ..
         }) = self.seq_delimiters.last()
         {
-            // item value
+            let len = match len.get() {
+                Some(len) => len as usize,
+                None => return Some(UndefinedItemLengthSnafu.fail()),
+            };
 
-            let len = len.get().expect("length should be explicit, error missing") as usize;
-            let mut value = vec![0; len];
+            if self.offset_table_next {
+                // offset table
+                let mut offset_table = Vec::with_capacity(len);
 
-            // need to pop item delimiter on the next iteration
-            self.delimiter_check_pending = true;
-            Some(
-                self.parser
-                    .read_bytes(&mut value[..])
-                    .map(|_| Ok(DataToken::ItemValue(value)))
-                    .unwrap_or_else(|e| Err(e).context(ReadItemValue { len: len as u32 })),
-            )
+                self.offset_table_next = false;
+
+                // need to pop item delimiter on the next iteration
+                self.delimiter_check_pending = true;
+
+                Some(
+                    match self.parser.read_u32_to_vec(len as u32, &mut offset_table) {
+                        Ok(()) => Ok(DataToken::OffsetTable(offset_table)),
+                        Err(e) => Err(e).context(ReadItemValueSnafu { len: len as u32 }),
+                    },
+                )
+            } else {
+                // item value
+                let mut value = Vec::with_capacity(len);
+
+                // need to pop item delimiter on the next iteration
+                self.delimiter_check_pending = true;
+                Some(
+                    self.parser
+                        .read_to_vec(len as u32, &mut value)
+                        .map(|_| Ok(DataToken::ItemValue(value)))
+                        .unwrap_or_else(|e| Err(e).context(ReadItemValueSnafu { len: len as u32 })),
+                )
+            }
         } else if let Some(header) = self.last_header {
             if header.is_encapsulated_pixeldata() {
                 self.push_sequence_token(SeqTokenType::Sequence, Length::UNDEFINED, true);
@@ -337,6 +359,8 @@ where
                             // items can be empty
                             if len == Length(0) {
                                 self.delimiter_check_pending = true;
+                            } else {
+                                self.offset_table_next = true;
                             }
                             Some(Ok(DataToken::ItemStart { len }))
                         }
@@ -348,12 +372,12 @@ where
                         }
                         item => {
                             self.hard_break = true;
-                            Some(UnexpectedItemTag { tag: item.tag() }.fail())
+                            Some(UnexpectedItemTagSnafu { tag: item.tag() }.fail())
                         }
                     },
                     Err(e) => {
                         self.hard_break = true;
-                        Some(Err(e).context(ReadItemHeader))
+                        Some(Err(e).context(ReadItemHeaderSnafu))
                     }
                 }
             } else {
@@ -397,6 +421,10 @@ where
                     ..
                 }) => {
                     self.in_sequence = true;
+                    // pop item delimiter
+                    self.seq_delimiters.pop();
+                    // sequences can end after this token
+                    self.delimiter_check_pending = true;
                     Some(Ok(DataToken::ItemEnd))
                 }
                 Ok(header) if header.is_encapsulated_pixeldata() => {
@@ -437,14 +465,14 @@ where
                 }
                 Err(e) => {
                     self.hard_break = true;
-                    Some(Err(e).context(ReadHeader))
+                    Some(Err(e).context(ReadHeaderSnafu))
                 }
             }
         }
     }
 }
 
-impl<S, D> DataSetReader<S, D>
+impl<S> DataSetReader<S>
 where
     S: StatefulDecode,
 {
@@ -452,7 +480,7 @@ where
         if let Some(sd) = self.seq_delimiters.last() {
             if let Some(len) = sd.len.get() {
                 let end_of_sequence = sd.base_offset + len as u64;
-                let bytes_read = self.parser.bytes_read();
+                let bytes_read = self.parser.position();
                 match end_of_sequence.cmp(&bytes_read) {
                     Ordering::Equal => {
                         // end of delimiter, as indicated by the element's length
@@ -471,7 +499,7 @@ where
                         return Ok(Some(token));
                     }
                     Ordering::Less => {
-                        return InconsistentSequenceEnd {
+                        return InconsistentSequenceEndSnafu {
                             end_of_sequence,
                             bytes_read,
                         }
@@ -491,7 +519,7 @@ where
             typ,
             pixel_data,
             len,
-            base_offset: self.parser.bytes_read(),
+            base_offset: self.parser.position(),
         })
     }
 
@@ -501,147 +529,25 @@ where
             ValueReadStrategy::Preserved => self.parser.read_value_preserved(header),
             ValueReadStrategy::Raw => self.parser.read_value_bytes(header),
         }
-        .context(ReadValue {
+        .context(ReadValueSnafu {
             len: header.len.0,
             tag: header.tag,
         })
     }
 }
 
-/// An iterator for retrieving DICOM object element markers from a random
-/// access data source.
-#[derive(Debug)]
-pub struct LazyDataSetReader<DS, P> {
-    parser: P,
-    depth: u32,
-    in_sequence: bool,
-    hard_break: bool,
-    phantom: PhantomData<DS>,
-}
-
-impl<'s> LazyDataSetReader<&'s mut dyn Read, DynStatefulDecoder<'s>> {
-    /// Create a new iterator with the given random access source,
-    /// while considering the given transfer syntax and specific character set.
-    pub fn new_with(
-        source: &'s mut dyn ReadSeek,
-        ts: &TransferSyntax,
-        cs: SpecificCharacterSet,
-    ) -> Result<Self> {
-        let parser = StatefulDecoder::new_with(source, ts, cs).context(CreateDecoder)?;
-
-        Ok(LazyDataSetReader {
-            parser,
-            depth: 0,
-            in_sequence: false,
-            hard_break: false,
-            phantom: PhantomData,
-        })
-    }
-}
-
-impl<DS, P> LazyDataSetReader<DS, P>
-where
-    P: StatefulDecode,
-{
-    /// Create a new iterator with the given parser.
-    pub fn new(parser: P) -> LazyDataSetReader<DS, P> {
-        LazyDataSetReader {
-            parser,
-            depth: 0,
-            in_sequence: false,
-            hard_break: false,
-            phantom: PhantomData,
-        }
-    }
-
-    fn create_element_marker(&mut self, header: DataElementHeader) -> Result<DicomElementMarker> {
-        Ok(DicomElementMarker {
-            header,
-            pos: self.parser.bytes_read(),
-        })
-    }
-
-    fn create_item_marker(&mut self, header: SequenceItemHeader) -> Result<DicomElementMarker> {
-        Ok(DicomElementMarker {
-            header: From::from(header),
-            pos: self.parser.bytes_read(),
-        })
-    }
-}
-
-impl<P> Iterator for LazyDataSetReader<(), P>
-where
-    P: StatefulDecode,
-{
-    type Item = Result<DicomElementMarker>;
-
-    fn next(&mut self) -> Option<Self::Item> {
-        if self.hard_break {
-            return None;
-        }
-        if self.in_sequence {
-            match self.parser.decode_item_header() {
-                Ok(header) => match header {
-                    header @ SequenceItemHeader::Item { .. } => {
-                        self.in_sequence = false;
-                        Some(self.create_item_marker(header))
-                    }
-                    SequenceItemHeader::ItemDelimiter => {
-                        self.in_sequence = true;
-                        Some(self.create_item_marker(header))
-                    }
-                    SequenceItemHeader::SequenceDelimiter => {
-                        self.depth -= 1;
-                        self.in_sequence = false;
-                        Some(self.create_item_marker(header))
-                    }
-                },
-                Err(e) => {
-                    self.hard_break = true;
-                    Some(Err(e).context(ReadItemHeader))
-                }
-            }
-        } else {
-            match self.parser.decode_header() {
-                Ok(
-                    header
-                    @
-                    DataElementHeader {
-                        tag: Tag(0x0008, 0x0005),
-                        ..
-                    },
-                ) => {
-                    let marker = self.create_element_marker(header);
-
-                    Some(marker)
-                }
-                Ok(header) => {
-                    // check if SQ
-                    if header.vr() == VR::SQ {
-                        self.in_sequence = true;
-                        self.depth += 1;
-                    }
-                    Some(self.create_element_marker(header))
-                }
-                Err(e) => {
-                    self.hard_break = true;
-                    Some(Err(e).context(ReadHeader))
-                }
-            }
-        }
-    }
-}
-
 #[cfg(test)]
 mod tests {
-    use super::{DataSetReader, DataToken, StatefulDecode, StatefulDecoder};
+    use super::{DataSetReader, DataToken, StatefulDecode};
+    use crate::stateful::decode::StatefulDecoder;
     use dicom_core::header::{DataElementHeader, Length};
     use dicom_core::value::PrimitiveValue;
     use dicom_core::{Tag, VR};
     use dicom_encoding::decode::basic::LittleEndianBasicDecoder;
-    use dicom_encoding::text::DefaultCharacterSetCodec;
-    use dicom_encoding::transfer_syntax::explicit_le::ExplicitVRLittleEndianDecoder;
-    use dicom_encoding::transfer_syntax::implicit_le::ImplicitVRLittleEndianDecoder;
+    use dicom_encoding::decode::{
+        explicit_le::ExplicitVRLittleEndianDecoder, implicit_le::ImplicitVRLittleEndianDecoder,
+    };
+    use dicom_encoding::text::SpecificCharacterSet;
 
     fn validate_dataset_reader_implicit_vr<I>(data: &[u8], ground_truth: I)
     where
@@ -652,7 +558,7 @@ mod tests {
             &mut cursor,
             ImplicitVRLittleEndianDecoder::default(),
             LittleEndianBasicDecoder::default(),
-            Box::new(DefaultCharacterSetCodec::default()) as Box<_>, // trait object
+            SpecificCharacterSet::Default,
         );
 
         validate_dataset_reader(data, parser, ground_truth)
@@ -667,7 +573,7 @@ mod tests {
             &mut cursor,
             ExplicitVRLittleEndianDecoder::default(),
             LittleEndianBasicDecoder::default(),
-            Box::new(DefaultCharacterSetCodec::default()) as Box<_>, // trait object
+            SpecificCharacterSet::Default,
         );
 
         validate_dataset_reader(&data, parser, ground_truth)
@@ -685,7 +591,11 @@ mod tests {
         while let Some((res, gt_token)) = iter.next() {
             let token = res.expect("should parse without an error");
             eprintln!("Next token: {:2?} ; Expected: {:2?}", token, gt_token);
-            assert_eq!(token, gt_token);
+            assert_eq!(
+                token, gt_token,
+                "Got token {:2?} ; but expected {:2?}",
+                token, gt_token
+            );
         }
 
         assert_eq!(
@@ -693,7 +603,11 @@ mod tests {
             0,            // we have already read all of them
             "unexpected number of tokens remaining"
         );
-        assert_eq!(dset_reader.parser.bytes_read(), data.len() as u64);
+        assert_eq!(
+            dset_reader.parser.position(),
+            data.len() as u64,
+            "Decoder position did not match end of data",
+        );
     }
 
     #[test]
@@ -1027,7 +941,7 @@ mod tests {
         let ground_truth = vec![
             DataToken::PixelSequenceStart,
             DataToken::ItemStart { len: Length(4) },
-            DataToken::ItemValue(vec![0x10, 0x00, 0x00, 0x00]),
+            DataToken::OffsetTable(vec![16]),
             DataToken::ItemEnd,
             DataToken::ItemStart { len: Length(32) },
             DataToken::ItemValue(vec![0x99; 32]),
@@ -1043,4 +957,104 @@ mod tests {
 
         validate_dataset_reader_explicit_vr(DATA, ground_truth);
     }
+
+    #[test]
+    fn read_dataset_in_dataset() {
+        #[rustfmt::skip]
+        const DATA: &'static [u8; 138] = &[
+            // 0: (2001, 9000) private sequence
+            0x01, 0x20, 0x00, 0x90, //
+            // length: undefined
+            0xFF, 0xFF, 0xFF, 0xFF, //
+            // 8: Item start
+            0xFE, 0xFF, 0x00, 0xE0, //
+            // Item length explicit (114 bytes)
+            0x72, 0x00, 0x00, 0x00, //
+            // 16: (0008,1115) ReferencedSeriesSequence
+            0x08, 0x00, 0x15, 0x11, //
+            // length: undefined
+            0xFF, 0xFF, 0xFF, 0xFF, //
+            // 24: Item start
+            0xFE, 0xFF, 0x00, 0xE0, //
+            // Item length undefined
+            0xFF, 0xFF, 0xFF, 0xFF, //
+            // 32: (0008,1140) ReferencedImageSequence
+            0x08, 0x00, 0x40, 0x11, //
+            // length: undefined
+            0xFF, 0xFF, 0xFF, 0xFF, //
+            // 40: Item start
+            0xFE, 0xFF, 0x00, 0xE0, //
+            // Item length undefined
+            0xFF, 0xFF, 0xFF, 0xFF, //
+            // 48: (0008,1150) ReferencedSOPClassUID
+            0x08, 0x00, 0x50, 0x11, //
+            // length: 26
+            0x1a, 0x00, 0x00, 0x00, //
+            // Value: "1.2.840.10008.5.1.4.1.1.7\0" (SecondaryCaptureImageStorage)
+            b'1', b'.', b'2', b'.', b'8', b'4', b'0', b'.', b'1', b'0', b'0', b'0', b'8', b'.',
+            b'5', b'.', b'1', b'.', b'4', b'.', b'1', b'.', b'1', b'.', b'7', b'\0',
+            // 82: Item End (ReferencedImageSequence)
+            0xFE, 0xFF, 0x0D, 0xE0, //
+            0x00, 0x00, 0x00, 0x00, //
+            // 90: Sequence End (ReferencedImageSequence)
+            0xFE, 0xFF, 0xDD, 0xE0, //
+            0x00, 0x00, 0x00, 0x00, //
+            // 98: Item End (ReferencedSeriesSequence)
+            0xFE, 0xFF, 0x0D, 0xE0, //
+            0x00, 0x00, 0x00, 0x00, //
+            // 106: Sequence End (ReferencedSeriesSequence)
+            0xFE, 0xFF, 0xDD, 0xE0, //
+            0x00, 0x00, 0x00, 0x00, //
+            // 114: (2050,0020) PresentationLUTShape (CS)
+            0x50, 0x20, 0x20, 0x00, //
+            // length: 8
+            0x08, 0x00, 0x00, 0x00, //
+            b'I', b'D', b'E', b'N', b'T', b'I', b'T', b'Y', //
+            // 130: Sequence end
+            0xFE, 0xFF, 0xDD, 0xE0, //
+            0x00, 0x00, 0x00, 0x00, //
+        ];
+
+        let ground_truth = vec![
+            DataToken::SequenceStart {
+                tag: Tag(0x2001, 0x9000),
+                len: Length::UNDEFINED,
+            },
+            DataToken::ItemStart { len: Length(114) },
+            DataToken::SequenceStart {
+                tag: Tag(0x0008, 0x1115),
+                len: Length::UNDEFINED,
+            },
+            DataToken::ItemStart {
+                len: Length::UNDEFINED,
+            },
+            DataToken::SequenceStart {
+                tag: Tag(0x0008, 0x1140),
+                len: Length::UNDEFINED,
+            },
+            DataToken::ItemStart {
+                len: Length::UNDEFINED,
+            },
+            DataToken::ElementHeader(DataElementHeader {
+                tag: Tag(0x0008, 0x1150),
+                vr: VR::UI,
+                len: Length(26),
+            }),
+            DataToken::PrimitiveValue(PrimitiveValue::from("1.2.840.10008.5.1.4.1.1.7\0")),
+            DataToken::ItemEnd,
+            DataToken::SequenceEnd,
+            DataToken::ItemEnd,
+            DataToken::SequenceEnd,
+            DataToken::ElementHeader(DataElementHeader {
+                tag: Tag(0x2050, 0x0020),
+                vr: VR::CS,
+                len: Length(8),
+            }),
+            DataToken::PrimitiveValue(PrimitiveValue::from("IDENTITY")),
+            DataToken::ItemEnd, // inserted automatically
+            DataToken::SequenceEnd,
+        ];
+
+        validate_dataset_reader_implicit_vr(DATA, ground_truth);
+    }
 }
