@@ -1,9 +1,16 @@
 //! Module for the data set writer
+//!
+//! This module contains a mid-level abstraction for printing DICOM data sets
+//! sequentially.
+//! The [`DataSetWriter`] receieves data tokens to be encoded and written
+//! to a writer.
+//! In this process, the writer will also adapt values
+//! to the necessary DICOM encoding rules.
 use crate::dataset::*;
 use crate::stateful::encode::StatefulEncoder;
 use dicom_core::{DataElementHeader, Length, VR};
 use dicom_encoding::encode::EncodeTo;
-use dicom_encoding::text::{SpecificCharacterSet, TextCodec};
+use dicom_encoding::text::SpecificCharacterSet;
 use dicom_encoding::transfer_syntax::DynEncoder;
 use dicom_encoding::TransferSyntax;
 use snafu::{Backtrace, OptionExt, ResultExt, Snafu};
@@ -78,30 +85,37 @@ struct SeqToken {
 /// This is analogous to the `DatasetReader` type for converting data
 /// set tokens to bytes.
 #[derive(Debug)]
-pub struct DataSetWriter<W, E, T> {
+pub struct DataSetWriter<W, E, T = SpecificCharacterSet> {
     printer: StatefulEncoder<W, E, T>,
     seq_tokens: Vec<SeqToken>,
     last_de: Option<DataElementHeader>,
 }
 
-impl<'w, W: 'w> DataSetWriter<W, DynEncoder<'w, W>, Box<dyn TextCodec>>
+impl<'w, W: 'w> DataSetWriter<W, DynEncoder<'w, W>>
 where
     W: Write,
 {
     pub fn with_ts_cs(to: W, ts: &TransferSyntax, charset: SpecificCharacterSet) -> Result<Self> {
-        let encoder = ts.encoder_for().context(UnsupportedTransferSyntax {
+        let encoder = ts.encoder_for().context(UnsupportedTransferSyntaxSnafu {
             ts_uid: ts.uid(),
             ts_alias: ts.name(),
         })?;
-        let text = charset
-            .codec()
-            .context(UnsupportedCharacterSet { charset })?;
-        Ok(DataSetWriter::new(to, encoder, text))
+        Ok(DataSetWriter::new_with_codec(to, encoder, charset))
+    }
+}
+
+impl<W, E> DataSetWriter<W, E> {
+    pub fn new(to: W, encoder: E) -> Self {
+        DataSetWriter {
+            printer: StatefulEncoder::new(to, encoder, SpecificCharacterSet::Default),
+            seq_tokens: Vec::new(),
+            last_de: None,
+        }
     }
 }
 
 impl<W, E, T> DataSetWriter<W, E, T> {
-    pub fn new(to: W, encoder: E, text: T) -> Self {
+    pub fn new_with_codec(to: W, encoder: E, text: T) -> Self {
         DataSetWriter {
             printer: StatefulEncoder::new(to, encoder, text),
             seq_tokens: Vec::new(),
@@ -110,11 +124,10 @@ impl<W, E, T> DataSetWriter<W, E, T> {
     }
 }
 
-impl<W, E, T> DataSetWriter<W, E, T>
+impl<W, E> DataSetWriter<W, E>
 where
     W: Write,
     E: EncodeTo<W>,
-    T: TextCodec,
 {
     /// Feed the given sequence of tokens which are part of the same data set.
     #[inline]
@@ -141,7 +154,7 @@ where
                     typ: SeqTokenType::Sequence,
                     len,
                 });
-                self.write_impl(token)?;
+                self.write_impl(&token)?;
                 Ok(())
             }
             DataToken::ItemStart { len } => {
@@ -149,14 +162,14 @@ where
                     typ: SeqTokenType::Item,
                     len,
                 });
-                self.write_impl(token)?;
+                self.write_impl(&token)?;
                 Ok(())
             }
             DataToken::ItemEnd => {
                 // only write if it's an unknown length item
                 if let Some(seq_start) = self.seq_tokens.pop() {
                     if seq_start.typ == SeqTokenType::Item && seq_start.len.is_undefined() {
-                        self.write_impl(token)?;
+                        self.write_impl(&token)?;
                     }
                 }
                 Ok(())
@@ -165,72 +178,81 @@ where
                 // only write if it's an unknown length sequence
                 if let Some(seq_start) = self.seq_tokens.pop() {
                     if seq_start.typ == SeqTokenType::Sequence && seq_start.len.is_undefined() {
-                        self.write_impl(token)?;
+                        self.write_impl(&token)?;
                     }
                 }
                 Ok(())
             }
             DataToken::ElementHeader(de) => {
+                // save the header for later
                 self.last_de = Some(de);
-                self.write_impl(token)
+
+                // postpone writing the header until the value token is given
+                Ok(())
             }
             token @ DataToken::PixelSequenceStart => {
                 self.seq_tokens.push(SeqToken {
                     typ: SeqTokenType::Sequence,
                     len: Length::UNDEFINED,
                 });
-                self.write_impl(token)
-            }
-            token @ DataToken::ItemValue(_) | token @ DataToken::PrimitiveValue(_) => {
-                self.write_impl(token)
+                self.write_impl(&token)
             }
+            token @ DataToken::ItemValue(_)
+            | token @ DataToken::PrimitiveValue(_)
+            | token @ DataToken::OffsetTable(_) => self.write_impl(&token),
         }
     }
 
-    fn write_impl(&mut self, token: DataToken) -> Result<()> {
+    fn write_impl(&mut self, token: &DataToken) -> Result<()> {
         match token {
             DataToken::ElementHeader(header) => {
                 self.printer
-                    .encode_element_header(header)
-                    .context(WriteHeader { tag: header.tag })?;
+                    .encode_element_header(*header)
+                    .context(WriteHeaderSnafu { tag: header.tag })?;
             }
             DataToken::SequenceStart { tag, len } => {
                 self.printer
-                    .encode_element_header(DataElementHeader::new(tag, VR::SQ, len))
-                    .context(WriteHeader { tag })?;
+                    .encode_element_header(DataElementHeader::new(*tag, VR::SQ, *len))
+                    .context(WriteHeaderSnafu { tag: *tag })?;
             }
             DataToken::PixelSequenceStart => {
                 let tag = Tag(0x7fe0, 0x0010);
                 self.printer
                     .encode_element_header(DataElementHeader::new(tag, VR::OB, Length::UNDEFINED))
-                    .context(WriteHeader { tag })?;
+                    .context(WriteHeaderSnafu { tag })?;
             }
             DataToken::SequenceEnd => {
                 self.printer
                     .encode_sequence_delimiter()
-                    .context(WriteSequenceDelimiter)?;
+                    .context(WriteSequenceDelimiterSnafu)?;
             }
             DataToken::ItemStart { len } => {
                 self.printer
                     .encode_item_header(len.0)
-                    .context(WriteItemHeader)?;
+                    .context(WriteItemHeaderSnafu)?;
             }
             DataToken::ItemEnd => {
                 self.printer
                     .encode_item_delimiter()
-                    .context(WriteItemDelimiter)?;
+                    .context(WriteItemDelimiterSnafu)?;
             }
             DataToken::PrimitiveValue(ref value) => {
-                let last_de = self.last_de.as_ref().with_context(|| UnexpectedToken {
+                let last_de = self.last_de.take().with_context(|| UnexpectedTokenSnafu {
                     token: token.clone(),
                 })?;
+
                 self.printer
-                    .encode_primitive(last_de, value)
-                    .context(WriteValue)?;
+                    .encode_primitive_element(&last_de, value)
+                    .context(WriteValueSnafu)?;
                 self.last_de = None;
             }
+            DataToken::OffsetTable(table) => {
+                self.printer
+                    .encode_offset_table(table)
+                    .context(WriteValueSnafu)?;
+            }
             DataToken::ItemValue(data) => {
-                self.printer.write_bytes(&data).context(WriteValue)?;
+                self.printer.write_bytes(data).context(WriteValueSnafu)?;
             }
         }
         Ok(())
@@ -246,9 +268,7 @@ mod tests {
         value::PrimitiveValue,
         Tag, VR,
     };
-    use dicom_encoding::encode::EncoderFor;
-    use dicom_encoding::text::DefaultCharacterSetCodec;
-    use dicom_encoding::transfer_syntax::explicit_le::ExplicitVRLittleEndianEncoder;
+    use dicom_encoding::encode::{explicit_le::ExplicitVRLittleEndianEncoder, EncoderFor};
 
     fn validate_dataset_writer<I>(tokens: I, ground_truth: &[u8])
     where
@@ -256,8 +276,7 @@ mod tests {
     {
         let mut raw_out: Vec<u8> = vec![];
         let encoder = EncoderFor::new(ExplicitVRLittleEndianEncoder::default());
-        let text = DefaultCharacterSetCodec::default();
-        let mut dset_writer = DataSetWriter::new(&mut raw_out, encoder, text);
+        let mut dset_writer = DataSetWriter::new(&mut raw_out, encoder);
 
         dset_writer.write_sequence(tokens).unwrap();
 
@@ -328,6 +347,44 @@ mod tests {
         validate_dataset_writer(tokens, GROUND_TRUTH);
     }
 
+    #[test]
+    fn write_element_overrides_len() {
+        let tokens = vec![
+            DataToken::ElementHeader(DataElementHeader {
+                // Specific Character Set (0008,0005)
+                tag: Tag(0x0008, 0x0005),
+                vr: VR::CS,
+                len: Length(10),
+            }),
+            DataToken::PrimitiveValue(PrimitiveValue::from("ISO_IR 100")),
+            DataToken::ElementHeader(DataElementHeader {
+                // Referring Physician's Name (0008,0090)
+                tag: Tag(0x0008, 0x0090),
+                vr: VR::PN,
+                // deliberately incorrect length
+                len: Length("Simões^João".len() as u32),
+            }),
+            DataToken::PrimitiveValue(PrimitiveValue::from("Simões^João")),
+        ];
+
+        #[rustfmt::skip]
+        static GROUND_TRUTH: &[u8] = &[
+            // Specific Character Set (0008,0005)
+            0x08, 0x00, 0x05, 0x00, //
+            b'C', b'S', // VR
+            0x0a, 0x00, // length: 10
+            b'I', b'S', b'O', b'_', b'I', b'R', b' ', b'1', b'0', b'0', // value = "ISO_IR 100"
+            // Referring Physician's Name (0008,0090)
+            0x08, 0x00, 0x90, 0x00, //
+            b'P', b'N', // VR
+            0x0c, 0x00, // length: 12
+            // value = "Simões^João "
+            b'S', b'i', b'm', 0xF5, b'e', b's', b'^', b'J', b'o', 0xE3, b'o', b' '
+        ];
+
+        validate_dataset_writer(tokens, GROUND_TRUTH);
+    }
+
     #[test]
     fn write_sequence_implicit() {
         let tokens = vec![
